@@ -113,6 +113,7 @@ typedef enum {
     S2N_ERR_CERT_TYPE_UNSUPPORTED,
     S2N_ERR_CERT_INVALID,
     S2N_ERR_CERT_MAX_CHAIN_DEPTH_EXCEEDED,
+    S2N_ERR_CERT_REJECTED,
     S2N_ERR_CRL_LOOKUP_FAILED,
     S2N_ERR_CRL_SIGNATURE,
     S2N_ERR_CRL_ISSUER,
@@ -312,6 +313,7 @@ typedef enum {
     S2N_ERR_KTLS_UNSUPPORTED_CONN,
     S2N_ERR_KTLS_ULP,
     S2N_ERR_KTLS_ENABLE_CRYPTO,
+    S2N_ERR_KTLS_BAD_CMSG,
     S2N_ERR_ATOMIC,
     S2N_ERR_T_USAGE_END,
 } s2n_error;
